--[[
	Table: ucl

	Holds all of the ucl variables and functions
]]
ULib.ucl = {}
local ucl = ULib.ucl -- Make it easier for us to refer to

-----------------------------------
--// Setup default information //--
-----------------------------------
ucl.users = {}
ucl.groups = {}
ucl.authed = {}
ucl.awaitingauth = {}
ucl.callbacks = {}

-------------------------------------------
--//Here's the meat of our control list//--
-------------------------------------------
--[[
	Function: ULib.ucl.query

	This function is used to see if a user has access to a command.

	Parameters:

		ply - The player to check access for
		command - The command you want to see if they have access to
		hide - *(Optional, defaults to false)* Normally, a listen host is automatically given access to everything. 
			Set this to true if you want to treat the listen host as a normal user.
]]
function ucl.query( ply, command, hide )
	if not ply:IsValid() or (not hide and ply:IsListenServerHost()) then return true end -- Grant full access to server host.

	local match -- This will hold the command matching ours
	if not command:find( " " ) then
		match = ULib.concommands[ command ]
	else
		local prefix_list = string.Explode( " ", command )
		local num = #prefix_list
		match = ULib.subconcommands

		for i=1, num do -- Get to the table list we need
			match = match[ prefix_list[ i ] ]
			if not match then
				break
			end
		end
	end

	if not match then -- Not under ULib control
		return ply:IsUserGroup( ULib.DEFAULT_ACCESS )
	end
	
	if not ucl.authed[ ply ] then
		return ply:IsUserGroup( match.access )
	end

	-- Now let's go through all applicable allows and denies for this user.
	local denies = { ucl.authed[ ply ].deny }
	local allows = { ucl.authed[ ply ].allow }
	-- We grabbed their user allow/deny, now let's grab all the group allow/denys.
	for _, v in pairs( ucl.authed[ ply ].groups ) do
		if ucl.groups[ v ] then
			table.insert( denies, ucl.groups[ v ].deny )
			table.insert( allows, ucl.groups[ v ].allow )			
		end
	end
	
	for _, v in pairs( denies ) do
		if table.HasValue( v, command ) then return false end
	end
	
	for _, v in pairs( allows ) do
		if table.HasValue( v, command ) then return true end
	end

	-- No specific instruction, fall back on command default.
	return ply:IsUserGroup( match.access )
end



------------------
--//Meta hooks//--
------------------
local meta = FindMetaTable( "Player" )
if not meta then return end

function meta:IsAdmin()
	if self:IsSuperAdmin() then return true end
	if self:IsUserGroup( "admin" ) then return true end

	return false
end

function meta:IsSuperAdmin()
	-- Always a super admin if this is our server
	if SERVER and self:IsListenServerHost() then return true end

	return self:IsUserGroup( "superadmin" )
end

function meta:IsUserGroup( name )
	if name == ULib.ACCESS_ALL then return true end -- Everyone has default access
	if name == ULib.ACCESS_NONE then return false end -- No one has this access

	if not ucl.authed[ self ] then return false end

	-- Let's work our way up the access chain.
	if name == ULib.ACCESS_OPERATOR then
		if table.HasValue( ucl.authed[ self ].groups, ULib.ACCESS_OPERATOR ) then return true end
		if table.HasValue( ucl.authed[ self ].groups, ULib.ACCESS_ADMIN ) then return true end
		if table.HasValue( ucl.authed[ self ].groups, ULib.ACCESS_SUPERADMIN ) then return true end
		return false
	end

	if name == ULib.ACCESS_ADMIN then
		if table.HasValue( ucl.authed[ self ].groups, ULib.ACCESS_ADMIN ) then return true end
		if table.HasValue( ucl.authed[ self ].groups, ULib.ACCESS_SUPERADMIN ) then return true end
		return false
	end

	if name == ULib.ACCESS_SUPERADMIN then
		if table.HasValue( ucl.authed[ self ].groups, ULib.ACCESS_SUPERADMIN ) then return true end
		return false
	end

	-- Else fall back on our default.
	return table.HasValue( ucl.authed[ self ].groups, name )
end

function meta:SetUserGroup( group )
	if not ULib.UCL_LOAD_DEFAULT then return end -- Ignore
	
	if table.HasValue( ucl.authed[ self ].groups ) then return end -- Already have it
	
	table.insert( ucl.authed[ self ].groups, group )
end

function meta:query( command, hide )
	return ULib.ucl.query( self, command, hide )
end

-----------------
--//Callbacks//--
-----------------

--[[
	Function: ULib.ucl.callCallbacks

	*Internal function, DO NOT CALL DIRECTLY.* This function calls access callbacks
	
	Parameters:
		
		ply - The player object.
]]
function ucl.callCallbacks( ply )
	for _, fn in ipairs( ucl.callbacks ) do
		local hasAccess = not not ucl.authed[ ply ] -- Double not! We want it in boolean form.
		pcall( fn, ply, hasAccess )
	end
end


--[[
	Function: ULib.ucl.addAccessCallback

	Pass a function and it will be called when someone receives access. Parameters it passes to the function are the player and an access bool.
	This function is useful for things like passing menus to users upon access. *This is also called when a user receives NO access*.
	
	Parameters:
		
		fn - The function to call.
]]
function ucl.addAccessCallback( fn )
	table.insert( ucl.callbacks, fn )
end

if CLIENT then -- Client section

	function ucl.rcvUserData( um )
		local ply = um:ReadEntity()
		ucl.authed[ ply ] = ULib.umsgRcv( um )
	end
	usermessage.Hook( "ULibUserUCL", ucl.rcvUserData )
	
	function ucl.rcvGroupData( um )
		local group = um:ReadString()
		ucl.groups[ group ] = ULib.umsgRcv( um )
	end
	usermessage.Hook( "ULibGroupUCL", ucl.rcvGroupData )
	
	function ucl.rcvFinished( um )
		ucl.callCallbacks( LocalPlayer() )
	end
	usermessage.Hook( "ULibFinishedUCL", ucl.rcvFinished )

end